﻿using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Optimization;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.Instances;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Common;
using HeuristicLab.Parameters;
using HeuristicLab.Data;
using System;
using System.Linq;
using HeuristicLab.Problems.PTSP.MoveEvaluators.OneShift;

namespace HeuristicLab.Problems.PTSP {
  [Item("Estimated Probabilistic Traveling Salesman Problem", "Represents an estimated Probabilistic Traveling Salesman Problem.")]
  [Creatable("Problems")]
  [StorableClass]
  public sealed class EstimatedProbabilisticTravelingSalesmanProblem : ProbabilisticTravelingSalesmanProblem, IStorableContent,
  IProblemInstanceConsumer<TSPData> {

    private ItemList<ItemList<IntValue>> realizations;

    [StorableConstructor]
    private EstimatedProbabilisticTravelingSalesmanProblem(bool deserializing) : base(deserializing) { }
    private EstimatedProbabilisticTravelingSalesmanProblem(EstimatedProbabilisticTravelingSalesmanProblem original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new EstimatedProbabilisticTravelingSalesmanProblem(this, cloner);
    }

    public override double Evaluate(Individual individual, IRandom random) {
      Permutation p = individual.Permutation();
      // Estimation-based evaluation
      Random r = new Random();
      double estimatedSum = 0;
      for (int i = 0; i < realizations.Count; i++) {
        int singleRealization = -1;
        for (int j = 0; j < realizations[i].Count; j++) {
          if (realizations[i][j].Value == 1) {
            if (singleRealization != -1) {
              estimatedSum += DistanceMatrix[singleRealization, p[j]];
            }
            singleRealization = p[j];
          }
        }
      }
      return estimatedSum / SampleSize.Value;
    }

    public EstimatedProbabilisticTravelingSalesmanProblem() {
      Parameters.Add(new ValueParameter<IntValue>("SampleSize", "Size of the sample for the estimation-based evaluation"));
      Operators.Add(new PTSPEstimatedInversionMovePathEvaluator());
      Operators.Add(new PTSPEstimatedInsertionEvaluator());
      Encoding.ConfigureOperators(Operators.OfType<IOperator>());
    }

    public override void Load(TSPData data) {
      base.Load(data);
      // For now uses sample size of 20 but should use Student's t-test
      SampleSize = new IntValue(20);
      realizations = new ItemList<ItemList<IntValue>>(SampleSize.Value);
      Random r = new Random();
      for (int i = 0; i < SampleSize.Value; i++) {
        realizations.Add(new ItemList<IntValue>());
        for (int j = 0; j < data.Dimension; j++) {
          if (ProbabilityMatrix[j].Value > r.NextDouble()) {
            realizations.ElementAt(i).Add(new IntValue(1));
          } else {
            realizations.ElementAt(i).Add(new IntValue(0));
          }
        }
        
      }
      foreach (var op in Operators.OfType<PTSPPathMoveEvaluator>()) {
        op.RealizationsParameter.Value = realizations;
      }
    }
  }
}